@@ -1,12 +1,13 @@
-import { render, screen } from '@testing-library/react';
+import { render, screen, waitFor } from '@testing-library/react';
 import { type Unsubscribable } from 'rxjs';
 
 import { dateTime, usePluginContext, PluginLoadingStrategy } from '@grafana/data';
 import { config, AppPluginConfig } from '@grafana/runtime';
-import appEvents from 'app/core/app_events';
+import { appEvents } from 'app/core/app_events';
 import { ShowModalReactEvent } from 'app/types/events';
 
 import { log } from './logs/log';
+import { resetLogMock } from './logs/testUtils';
 import {
   deepFreeze,
   handleErrorsInFn,
@@ -19,14 +20,34 @@ import {
   getAppPluginIdFromExposedComponentId,
   getAppPluginDependencies,
   getExtensionPointPluginMeta,
+  getMutationObserverProxy,
+  writableProxy,
+  isMutationObserverProxy,
 } from './utils';
 
 jest.mock('app/features/plugins/pluginSettings', () => ({
   ...jest.requireActual('app/features/plugins/pluginSettings'),
-  getPluginSettings: () => Promise.resolve({ info: { version: '1.0.0' } }),
+  getPluginSettings: () => Promise.resolve({ info: { version: '1.0.0' }, id: 'test-plugin' }),
 }));
 
 describe('Plugin Extensions / Utils', () => {
+  const originalEnv = config.buildInfo.env;
+
+  beforeEach(() => {
+    jest.spyOn(log, 'error').mockImplementation(() => {});
+    jest.spyOn(log, 'warning').mockImplementation(() => {});
+    jest.spyOn(log, 'debug').mockImplementation(() => {});
+    jest.spyOn(log, 'info').mockImplementation(() => {});
+    jest.spyOn(log, 'trace').mockImplementation(() => {});
+    jest.spyOn(log, 'fatal').mockImplementation(() => {});
+  });
+
+  afterEach(() => {
+    jest.resetAllMocks();
+
+    config.buildInfo.env = originalEnv;
+  });
+
   describe('deepFreeze()', () => {
     test('should not fail when called with primitive values', () => {
       // Although the type system doesn't allow to call it with primitive values, it can happen that the plugin just ignores these errors.
@@ -228,7 +249,7 @@ describe('Plugin Extensions / Utils', () => {
 
       expect(() => {
         proxy.a = 'b';
-      }).toThrowError(TypeError);
+      }).toThrow(TypeError);
     });
 
     it('should not be possible to modify values in proxied array', () => {
@@ -236,7 +257,7 @@ describe('Plugin Extensions / Utils', () => {
 
       expect(() => {
         proxy[0] = 2;
-      }).toThrowError(TypeError);
+      }).toThrow(TypeError);
     });
 
     it('should not be possible to modify nested objects in proxied object', () => {
@@ -249,7 +270,58 @@ describe('Plugin Extensions / Utils', () => {
 
       expect(() => {
         proxy.a.c = 'testing';
-      }).toThrowError(TypeError);
+      }).toThrow(TypeError);
+    });
+
+    // This is to record what we are not able to do currently.
+    // (Due to Proxy.get() invariants limitations: https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Proxy/Proxy/get#invariants)
+    it('should not work with any objects that are already frozen', () => {
+      const obj = {
+        a: {
+          b: {
+            c: {
+              d: 'd',
+            },
+          },
+        },
+      };
+
+      Object.freeze(obj);
+      Object.freeze(obj.a);
+      Object.freeze(obj.a.b);
+
+      const proxy = getReadOnlyProxy(obj);
+
+      expect(() => {
+        proxy.a.b.c.d = 'testing';
+      }).toThrow(
+        "'get' on proxy: property 'a' is a read-only and non-configurable data property on the proxy target but the proxy did not return its actual value (expected '#<Object>' but got '#<Object>')"
+      );
+
+      expect(obj.a.b.c.d).toBe('d');
+    });
+
+    it('should throw a TypeError if a proxied object is trying to be frozen', () => {
+      const obj = {
+        a: {
+          b: {
+            c: {
+              d: 'd',
+            },
+          },
+        },
+      };
+
+      const proxy = getReadOnlyProxy(obj);
+
+      expect(() => Object.freeze(proxy)).toThrow(TypeError);
+      expect(() => Object.freeze(proxy.a)).toThrow(TypeError);
+      expect(() => Object.freeze(proxy.a.b)).toThrow(TypeError);
+
+      // Check if the original object is not frozen
+      expect(Object.isFrozen(obj)).toBe(false);
+      expect(Object.isFrozen(obj.a)).toBe(false);
+      expect(Object.isFrozen(obj.a.b)).toBe(false);
     });
 
     it('should not be possible to modify nested arrays in proxied object', () => {
@@ -262,7 +334,7 @@ describe('Plugin Extensions / Utils', () => {
 
       expect(() => {
         proxy.a.c[0] = 'testing';
-      }).toThrowError(TypeError);
+      }).toThrow(TypeError);
     });
 
     it('should be possible to modify source object', () => {
@@ -320,6 +392,235 @@ describe('Plugin Extensions / Utils', () => {
     });
   });
 
+  describe('getMutationObserverProxy()', () => {
+    describe('in development mode', () => {
+      beforeEach(() => {
+        config.buildInfo.env = 'development';
+      });
+
+      it('should be possible to modify values in proxied object, but logs an error', () => {
+        const proxy = getMutationObserverProxy(
+          { a: 'a' },
+          { pluginId: 'myorg-cool-datasource', source: 'datasource', pluginVersion: '1.2.3' }
+        );
+
+        expect(() => {
+          proxy.a = 'b';
+        }).not.toThrow();
+
+        expect(log.error).toHaveBeenCalledWith(
+          `Attempted to mutate object property "a" from datasource with id myorg-cool-datasource and version 1.2.3`,
+          {
+            stack: expect.any(String),
+          }
+        );
+
+        expect(proxy.a).toBe('b');
+      });
+
+      it('should be possible to call defineProperty, but logs a debug message', () => {
+        const obj: { a: string; b?: string } = { a: 'a' };
+        const proxy = getMutationObserverProxy(obj, { pluginId: 'myorg-cool-extension' });
+
+        expect(() => {
+          Object.defineProperty(proxy, 'b', {
+            value: 'b',
+            writable: false,
+          });
+        }).not.toThrow();
+
+        expect(log.debug).toHaveBeenCalledWith(
+          `Attempted to define object property "b" from extension with id myorg-cool-extension and version unknown`,
+          {
+            stack: expect.any(String),
+          }
+        );
+
+        expect(proxy.b).toBe('b');
+      });
+
+      it('should be possible to delete properties, but logs an error', () => {
+        const proxy = getMutationObserverProxy({
+          a: {
+            c: 'c',
+          },
+          b: 'b',
+        });
+
+        expect(() => {
+          // @ts-ignore - This is to test the logic
+          delete proxy.a.c;
+        }).not.toThrow();
+
+        expect(log.error).toHaveBeenCalledWith(
+          `Attempted to delete object property "c" from extension with id unknown and version unknown`,
+          {
+            stack: expect.any(String),
+          }
+        );
+
+        expect(proxy.a.c).toBeUndefined();
+      });
+    });
+
+    describe('in production mode', () => {
+      beforeEach(() => {
+        config.buildInfo.env = 'production';
+      });
+
+      it('should be possible to modify values in proxied object, but logs a warning', () => {
+        const proxy = getMutationObserverProxy({ a: 'a' }, { pluginId: 'myorg-cool-datasource', source: 'datasource' });
+
+        expect(() => {
+          proxy.a = 'b';
+        }).not.toThrow();
+
+        expect(log.warning).toHaveBeenCalledWith(
+          `Attempted to mutate object property "a" from datasource with id myorg-cool-datasource and version unknown`,
+          {
+            stack: expect.any(String),
+          }
+        );
+
+        expect(proxy.a).toBe('b');
+      });
+
+      it('should be possible to call defineProperty, but logs a debug message', () => {
+        const obj: { a: string; b?: string } = { a: 'a' };
+        const proxy = getMutationObserverProxy(obj, { pluginId: 'myorg-cool-extension' });
+
+        expect(() => {
+          Object.defineProperty(proxy, 'b', {
+            value: 'b',
+            writable: false,
+          });
+        }).not.toThrow();
+
+        expect(log.debug).toHaveBeenCalledWith(
+          `Attempted to define object property "b" from extension with id myorg-cool-extension and version unknown`,
+          {
+            stack: expect.any(String),
+          }
+        );
+
+        expect(proxy.b).toBe('b');
+      });
+
+      it('should be possible to delete properties, but logs a warning', () => {
+        const proxy = getMutationObserverProxy({
+          a: {
+            c: 'c',
+          },
+          b: 'b',
+        });
+
+        expect(() => {
+          // @ts-ignore - This is to test the logic
+          delete proxy.a.c;
+        }).not.toThrow();
+
+        expect(log.warning).toHaveBeenCalledWith(
+          `Attempted to delete object property "c" from extension with id unknown and version unknown`,
+          {
+            stack: expect.any(String),
+          }
+        );
+
+        expect(proxy.a.c).toBeUndefined();
+      });
+    });
+  });
+
+  describe('writableProxy()', () => {
+    const originalEnv = config.buildInfo.env;
+
+    afterEach(() => {
+      config.buildInfo.env = originalEnv;
+    });
+
+    it('should return the same value for primitive types', () => {
+      expect(writableProxy(1)).toBe(1);
+      expect(writableProxy('a')).toBe('a');
+      expect(writableProxy(true)).toBe(true);
+      expect(writableProxy(false)).toBe(false);
+      expect(writableProxy(null)).toBe(null);
+      expect(writableProxy(undefined)).toBe(undefined);
+    });
+
+    it('should return a writable deep-copy of the original object in dev mode', () => {
+      config.buildInfo.env = 'development';
+
+      const obj = { a: 'a' };
+      const copy = writableProxy(obj, {
+        source: 'datasource',
+        pluginId: 'myorg-cool-datasource',
+        pluginVersion: '1.2.3',
+      });
+
+      expect(copy).not.toBe(obj);
+      expect(copy.a).toBe('a');
+      expect(isMutationObserverProxy(copy)).toBe(true);
+      expect(() => {
+        copy.a = 'b';
+      }).not.toThrow();
+
+      expect(log.error).toHaveBeenCalledWith(
+        `Attempted to mutate object property "a" from datasource with id myorg-cool-datasource and version 1.2.3`,
+        {
+          stack: expect.any(String),
+        }
+      );
+
+      expect(copy.a).toBe('b');
+    });
+
+    it('should return a writable deep-copy of the original object in production mode', () => {
+      config.buildInfo.env = 'production';
+
+      const obj = { a: 'a' };
+      const copy = writableProxy(obj, { source: 'datasource', pluginId: 'myorg-cool-datasource' });
+
+      expect(copy).not.toBe(obj);
+      expect(copy.a).toBe('a');
+      expect(isMutationObserverProxy(copy)).toBe(true);
+      expect(() => {
+        copy.a = 'b';
+      }).not.toThrow();
+
+      expect(log.warning).toHaveBeenCalledWith(
+        `Attempted to mutate object property "a" from datasource with id myorg-cool-datasource and version unknown`,
+        {
+          stack: expect.any(String),
+        }
+      );
+
+      expect(copy.a).toBe('b');
+    });
+
+    it('should allow freezing the object in production mode', () => {
+      config.buildInfo.env = 'production';
+
+      const obj = { a: 'a', b: { c: 'c' } };
+      const copy = writableProxy(obj);
+
+      expect(() => {
+        Object.freeze(copy);
+        Object.freeze(copy.b);
+      }).not.toThrow();
+
+      expect(Object.isFrozen(copy)).toBe(true);
+      expect(Object.isFrozen(copy.b)).toBe(true);
+      expect(copy.b).toEqual({ c: 'c' });
+
+      expect(log.debug).toHaveBeenCalledWith(
+        `Attempted to define object property "a" from extension with id unknown and version unknown`,
+        {
+          stack: expect.any(String),
+        }
+      );
+    });
+  });
+
   describe('createOpenModalFunction()', () => {
     let renderModalSubscription: Unsubscribable | undefined;
 
@@ -337,7 +638,11 @@ describe('Plugin Extensions / Utils', () => {
 
     it('should open modal with provided title and body', async () => {
       const pluginId = 'grafana-worldmap-panel';
-      const openModal = createOpenModalFunction(pluginId);
+      const openModal = createOpenModalFunction({
+        pluginId,
+        extensionPointId: 'myorg-extensions-app/link/v1',
+        title: 'Title in modal',
+      });
 
       openModal({
         title: 'Title in modal',
@@ -351,7 +656,11 @@ describe('Plugin Extensions / Utils', () => {
 
     it('should open modal with default width if not specified', async () => {
       const pluginId = 'grafana-worldmap-panel';
-      const openModal = createOpenModalFunction(pluginId);
+      const openModal = createOpenModalFunction({
+        pluginId,
+        extensionPointId: 'myorg-extensions-app/link/v1',
+        title: 'Title in modal',
+      });
 
       openModal({
         title: 'Title in modal',
@@ -367,7 +676,11 @@ describe('Plugin Extensions / Utils', () => {
 
     it('should open modal with specified width', async () => {
       const pluginId = 'grafana-worldmap-panel';
-      const openModal = createOpenModalFunction(pluginId);
+      const openModal = createOpenModalFunction({
+        pluginId,
+        extensionPointId: 'myorg-extensions-app/link/v1',
+        title: 'Title in modal',
+      });
 
       openModal({
         title: 'Title in modal',
@@ -383,7 +696,11 @@ describe('Plugin Extensions / Utils', () => {
 
     it('should open modal with specified height', async () => {
       const pluginId = 'grafana-worldmap-panel';
-      const openModal = createOpenModalFunction(pluginId);
+      const openModal = createOpenModalFunction({
+        pluginId,
+        extensionPointId: 'myorg-extensions-app/link/v1',
+        title: 'Title in modal',
+      });
 
       openModal({
         title: 'Title in modal',
@@ -399,7 +716,11 @@ describe('Plugin Extensions / Utils', () => {
 
     it('should open modal with the plugin context being available', async () => {
       const pluginId = 'grafana-worldmap-panel';
-      const openModal = createOpenModalFunction(pluginId);
+      const openModal = createOpenModalFunction({
+        pluginId,
+        extensionPointId: 'myorg-extensions-app/link/v1',
+        title: 'Title in modal',
+      });
 
       const ModalContent = () => {
         const context = usePluginContext();
@@ -415,17 +736,111 @@ describe('Plugin Extensions / Utils', () => {
       const modal = await screen.findByRole('dialog');
       expect(modal).toHaveTextContent('Version: 1.0.0');
     });
+
+    it('should add a wrapper div with a "data-plugin-sandbox" attribute', async () => {
+      const pluginId = 'grafana-worldmap-panel';
+      const openModal = createOpenModalFunction({
+        pluginId,
+        extensionPointId: 'myorg-extensions-app/link/v1',
+        title: 'Title in modal',
+      });
+
+      openModal({
+        title: 'Title in modal',
+        body: () => <div>Text in body</div>,
+      });
+
+      expect(await screen.findByRole('dialog')).toBeVisible();
+
+      expect(screen.getByTestId('plugin-sandbox-wrapper')).toHaveAttribute(
+        'data-plugin-sandbox',
+        'grafana-worldmap-panel'
+      );
+    });
+
+    it('should show an error alert in the modal IN DEV MODE if the extension throws an error', async () => {
+      config.buildInfo.env = 'development';
+      jest.spyOn(console, 'error').mockImplementation(() => {});
+
+      const pluginId = 'grafana-worldmap-panel';
+      const extensionTitle = 'Title in modal';
+      const openModal = createOpenModalFunction({
+        pluginId,
+        extensionPointId: 'myorg-extensions-app/link/v1',
+        title: extensionTitle,
+      });
+
+      const ModalContent = () => {
+        throw new Error('Test error');
+      };
+
+      openModal({
+        title: extensionTitle,
+        body: ModalContent,
+      });
+
+      await screen.findByRole('dialog');
+
+      expect(log.error).toHaveBeenCalledTimes(1);
+      expect(log.error).toHaveBeenCalledWith(`Extension "${pluginId}/${extensionTitle}" failed to load.`, {
+        message: 'Test error',
+        componentStack: expect.any(String),
+      });
+
+      expect(screen.getByText(`Extension failed to load: "${pluginId}/${extensionTitle}"`)).toBeVisible();
+    });
+
+    it('should also show an error alert in the modal IN PRODUCTION MODE if the extension throws an error', async () => {
+      config.buildInfo.env = 'production';
+      jest.spyOn(console, 'error').mockImplementation(() => {});
+
+      const pluginId = 'grafana-worldmap-panel';
+      const extensionTitle = 'Title in modal';
+      const openModal = createOpenModalFunction({
+        pluginId,
+        extensionPointId: 'myorg-extensions-app/link/v1',
+        title: extensionTitle,
+      });
+
+      const ModalContent = () => {
+        throw new Error('Test error');
+      };
+
+      openModal({
+        title: extensionTitle,
+        body: ModalContent,
+      });
+
+      await screen.findByRole('dialog');
+
+      expect(log.error).toHaveBeenCalledTimes(1);
+      expect(log.error).toHaveBeenCalledWith(`Extension "${pluginId}/${extensionTitle}" failed to load.`, {
+        message: 'Test error',
+        componentStack: expect.any(String),
+      });
+
+      expect(screen.getByText(`Extension failed to load: "${pluginId}/${extensionTitle}"`)).toBeVisible();
+    });
   });
 
-  describe('wrapExtensionComponentWithContext()', () => {
+  describe('wrapWithPluginContext()', () => {
     type ExampleComponentProps = {
-      audience?: string;
+      a: {
+        b: {
+          c: string;
+        };
+      };
+      override?: boolean;
     };
 
     const ExampleComponent = (props: ExampleComponentProps) => {
       const pluginContext = usePluginContext();
 
-      const audience = props.audience || 'Grafana';
+      const audience = props.a.b.c || 'Grafana';
+
+      if (props.override) {
+        props.a.b.c = 'OVERRIDE';
+      }
 
       return (
         <div>
@@ -434,11 +849,20 @@ describe('Plugin Extensions / Utils', () => {
       );
     };
 
+    beforeEach(() => {
+      resetLogMock(log);
+    });
+
     it('should make the plugin context available for the wrapped component', async () => {
       const pluginId = 'grafana-worldmap-panel';
-      const Component = wrapWithPluginContext(pluginId, ExampleComponent, log);
+      const Component = wrapWithPluginContext({
+        pluginId,
+        extensionTitle: 'ExampleComponent',
+        Component: ExampleComponent,
+        log,
+      });
 
-      render(<Component />);
+      render(<Component a={{ b: { c: 'Grafana' } }} />);
 
       expect(await screen.findByText('Hello Grafana!')).toBeVisible();
       expect(screen.getByText('Version: 1.0.0')).toBeVisible();
@@ -446,13 +870,130 @@ describe('Plugin Extensions / Utils', () => {
 
     it('should pass the properties into the wrapped component', async () => {
       const pluginId = 'grafana-worldmap-panel';
-      const Component = wrapWithPluginContext(pluginId, ExampleComponent, log);
+      const Component = wrapWithPluginContext({
+        pluginId,
+        extensionTitle: 'ExampleComponent',
+        Component: ExampleComponent,
+        log,
+      });
 
-      render(<Component audience="folks" />);
+      render(<Component a={{ b: { c: 'Grafana' } }} />);
 
-      expect(await screen.findByText('Hello folks!')).toBeVisible();
+      expect(await screen.findByText('Hello Grafana!')).toBeVisible();
       expect(screen.getByText('Version: 1.0.0')).toBeVisible();
     });
+
+    it('should not be possible to mutate the props in development mode, but it logs an error', async () => {
+      config.buildInfo.env = 'development';
+      const pluginId = 'grafana-worldmap-panel';
+      const Component = wrapWithPluginContext({
+        pluginId,
+        extensionTitle: 'ExampleComponent',
+        Component: ExampleComponent,
+        log,
+      });
+      const props = { a: { b: { c: 'Grafana' } } };
+
+      render(<Component {...props} override />);
+
+      expect(await screen.findByText('Hello Grafana!')).toBeVisible();
+
+      // Logs a warning
+      expect(log.error).toHaveBeenCalledTimes(1);
+      expect(log.error).toHaveBeenCalledWith(
+        `Attempted to mutate object property "c" from extension with id grafana-worldmap-panel and version 1.0.0`,
+        {
+          stack: expect.any(String),
+        }
+      );
+
+      // Not able to mutate the props in dev mode either
+      expect(props.a.b.c).toBe('Grafana');
+    });
+
+    it('should not be possible to mutate the props in production mode either, but it logs a warning', async () => {
+      config.buildInfo.env = 'production';
+      const pluginId = 'grafana-worldmap-panel';
+      const Component = wrapWithPluginContext({
+        pluginId,
+        extensionTitle: 'ExampleComponent',
+        Component: ExampleComponent,
+        log,
+      });
+      const props = { a: { b: { c: 'Grafana' } } };
+
+      render(<Component {...props} override />);
+
+      expect(await screen.findByText('Hello Grafana!')).toBeVisible();
+
+      // Logs a warning
+      expect(log.warning).toHaveBeenCalledTimes(1);
+      expect(log.warning).toHaveBeenCalledWith(
+        `Attempted to mutate object property "c" from extension with id grafana-worldmap-panel and version 1.0.0`,
+        {
+          stack: expect.any(String),
+        }
+      );
+
+      // Not able to mutate the props in production mode either
+      expect(props.a.b.c).toBe('Grafana');
+    });
+
+    it('should render an error alert IN DEV MODE if the extension throws an error', async () => {
+      config.buildInfo.env = 'development';
+      jest.spyOn(console, 'error').mockImplementation(() => {});
+
+      const pluginId = 'grafana-worldmap-panel';
+      const ComponentWithError = () => {
+        throw new Error('Test error');
+      };
+      const extensionTitle = 'ComponentWithError';
+      const WrappedComponent = wrapWithPluginContext({
+        pluginId,
+        extensionTitle,
+        Component: ComponentWithError,
+        log,
+      });
+
+      render(<WrappedComponent />);
+
+      expect(await screen.findByText(`Extension failed to load: "${pluginId}/${extensionTitle}"`)).toBeVisible();
+
+      expect(log.error).toHaveBeenCalledTimes(1);
+      expect(log.error).toHaveBeenCalledWith(`Extension "${pluginId}/${extensionTitle}" failed to load.`, {
+        message: 'Test error',
+        componentStack: expect.any(String),
+      });
+    });
+
+    it('should not render anything IN PRODUCTION MODE if the extension throws an error, but still logs an error', async () => {
+      config.buildInfo.env = 'production';
+      jest.spyOn(console, 'error').mockImplementation(() => {});
+
+      const pluginId = 'grafana-worldmap-panel';
+      const ComponentWithError = () => {
+        throw new Error('Test error');
+      };
+      const extensionTitle = 'ComponentWithError';
+      const WrappedComponent = wrapWithPluginContext({
+        pluginId,
+        extensionTitle,
+        Component: ComponentWithError,
+        log,
+      });
+
+      render(<WrappedComponent />);
+
+      await waitFor(() =>
+        expect(screen.queryByText(`Extension failed to load: "${pluginId}/${extensionTitle}"`)).not.toBeInTheDocument()
+      );
+
+      expect(log.error).toHaveBeenCalledTimes(1);
+      expect(log.error).toHaveBeenCalledWith(`Extension "${pluginId}/${extensionTitle}" failed to load.`, {
+        message: 'Test error',
+        componentStack: expect.any(String),
+      });
+    });
   });
 
   describe('getAppPluginConfigs()', () => {
